"""A CLI tool for checking disk space."""

import enum
import os
import platform
import shutil

from diskspaced.writer import Writer
from diskspaced.json_writer import JSONWriter
from diskspaced.grand_perspective_writer import GrandPerspectiveWriter


if platform.system() == "Windows":
    raise NotImplementedError("Windows is not supported by this tool.")


class OutputFormat(enum.Enum):
    """Represents the output format for the scan results."""

    JSON = "json"
    GRAND_PERSPECTIVE = "grandperspective"


def _get_block_size(path: str) -> int:
    """Get the block size for the given path.

    :param path: The path to get the block size for
    :returns: The block size
    """

    statvfs = os.statvfs(path)

    if platform.system() == "Darwin":
        return statvfs.f_frsize

    if platform.system() == "Linux":
        return statvfs.f_bsize

    raise NotImplementedError(f"Unsupported platform: {platform.system()}")


def _scan(folder_path: str, writer: Writer, process_in_order: bool) -> None:

    if os.path.isdir(folder_path):
        folder_details = os.stat(folder_path)

        writer.write_folder_start(
            os.path.basename(folder_path),
            int(folder_details.st_atime),
            int(folder_details.st_mtime),
            int(folder_details.st_ctime),
        )

    files = []
    folders = []

    for item in os.listdir(folder_path):
        full_path = os.path.join(folder_path, item)

        if os.path.isdir(full_path):
            folders.append(full_path)
        else:
            files.append((full_path, item))

    if process_in_order:
        folders.sort()
        files.sort(key=lambda x: x[1])

    for folder in folders:
        _scan(folder, writer, process_in_order)

    for file_path, file_name in files:
        file_details = os.stat(file_path)
        writer.write_file(
            file_name,
            file_details.st_size,
            int(file_details.st_atime),
            int(file_details.st_mtime),
            int(file_details.st_ctime),
        )

    writer.write_folder_end()


def scan(
    folder_path: str, output_path: str, output_format: OutputFormat, file_print_count: int
) -> None:
    """Scan the folder and write the results to the output path.

    :param folder_path: The path to scan
    :param output_path: The path to write the results to
    :param output_format: The format to write the results in
    :param file_print_count: The number of files to print after. Zero disables printing.
    """

    writer: Writer

    if output_format == OutputFormat.JSON:
        writer = JSONWriter(output_path, file_print_count)
    elif output_format == OutputFormat.GRAND_PERSPECTIVE:
        writer = GrandPerspectiveWriter(output_path, file_print_count)
    else:
        raise ValueError(f"Unknown output format: {output_format}")

    disk_usage = shutil.disk_usage(folder_path)
    block_size = _get_block_size(folder_path)

    writer.write_start(folder_path, disk_usage.total, disk_usage.used, disk_usage.free, block_size)

    _scan(folder_path, writer, file_print_count != 0)

    writer.write_end()



"""A CLI tool for checking disk space."""

import datetime
import os

from diskspaced import writer

# Example:
# <?xml version="1.0" encoding="UTF-8"?>
# <GrandPerspectiveScanDump appVersion="4" formatVersion="7">
#    <ScanInfo volumePath="/" volumeSize="994662584320" freeSpace="319055134720" scanTime="2024-08-23T13:58:23Z" fileSizeMeasure="physical">
#        <Folder name="Users/dalemyers/Projects/diskspaced/tests/data/test_scan_downloads_gp" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:56:53Z" accessed="2024-08-23T13:56:53Z">
#            <Folder name="a" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:56:53Z" accessed="2024-08-23T13:56:53Z">
#                <File name="one.txt" size="4096" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:58:13Z" accessed="2024-08-23T13:58:14Z" />
#                <Folder name="b" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:56:53Z" accessed="2024-08-23T13:56:53Z">
#                    <File name="two.txt" size="4096" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:58:15Z" accessed="2024-08-23T13:58:17Z" />
#                    <Folder name="c" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:56:53Z" accessed="2024-08-23T13:56:53Z">
#                        <File name="three.txt" size="4096" created="2024-08-23T13:56:53Z" modified="2024-08-23T13:58:20Z" accessed="2024-08-23T13:58:21Z" />
#                    </Folder>
#                </Folder>
#            </Folder>
#        </Folder>
#    </ScanInfo>
# </GrandPerspectiveScanDump>


class GrandPerspectiveWriter(writer.Writer):
    """A write out results as JSON."""

    DATE_FORMAT = "%Y-%m-%dT%H:%M:%SZ"

    FOLDER_OPEN = "<Folder ".encode("utf-8")
    FOLDER_CLOSE = "</Folder>\n".encode("utf-8")
    FILE_OPEN = "<File ".encode("utf-8")
    FILE_CLOSE = "/>\n".encode("utf-8")
    CLOSE_TAG = ">\n".encode("utf-8")
    ATTR_CLOSE = '" '.encode("utf-8")
    ATTR_NAME = 'name="'.encode("utf-8")
    ATTR_SIZE = 'size="'.encode("utf-8")
    ATTR_CREATED = 'created="'.encode("utf-8")
    ATTR_MODIFIED = 'modified="'.encode("utf-8")
    ATTR_ACCESSED = 'accessed="'.encode("utf-8")

    def write_start(
        self,
        root_path: str,
        disk_usage_total: int,
        disk_usage_used: int,
        disk_usage_free: int,
        block_size: int,
    ) -> None:
        """Write the start of the output file."""
        super().write_start(
            root_path, disk_usage_total, disk_usage_used, disk_usage_free, block_size
        )

        # pylint: disable=consider-using-with
        self.block_size = block_size
        self.file = open(self.output_path, "wb")
        # pylint: enable=consider-using-with

        if not root_path.endswith("/"):
            root_path += "/"

        now = datetime.datetime.now()
        scan_time = now.strftime(GrandPerspectiveWriter.DATE_FORMAT)

        # This only happens once, so we don't bother caching the encoded values
        self.file.write('<?xml version="1.0" encoding="UTF-8"?>\n'.encode("utf-8"))
        self.file.write(
            '  <GrandPerspectiveScanDump appVersion="4" formatVersion="7">\n'.encode("utf-8")
        )
        self.file.write(
            f'    <ScanInfo volumePath="{root_path}" volumeSize="{disk_usage_total}" freeSpace="{disk_usage_free}" scanTime="{scan_time}" fileSizeMeasure="physical">\n'.encode(
                "utf-8"
            )
        )

        if os.environ.get("PYTEST_CURRENT_TEST"):
            self.file.flush()

    def write_end(self) -> None:
        """Write the end of the output file."""
        super().write_end()
        self.file.write("  </ScanInfo>\n</GrandPerspectiveScanDump>".encode("utf-8"))
        self.file.close()

    def write_folder_start(
        self, folder_name: str, accessed_time: int, modified_time: int, created_time: int
    ) -> None:
        """Write the start of a folder entry."""

        super().write_folder_start(folder_name, accessed_time, modified_time, created_time)

        self.file.write(GrandPerspectiveWriter.FOLDER_OPEN)

        self.file.write(GrandPerspectiveWriter.ATTR_NAME)
        self.file.write(folder_name.encode("utf-8"))
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_CREATED)
        self.file.write(
            datetime.datetime.fromtimestamp(created_time)
            .strftime(GrandPerspectiveWriter.DATE_FORMAT)
            .encode("utf-8")
        )
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_MODIFIED)
        self.file.write(
            datetime.datetime.fromtimestamp(modified_time)
            .strftime(GrandPerspectiveWriter.DATE_FORMAT)
            .encode("utf-8")
        )
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_ACCESSED)
        self.file.write(
            datetime.datetime.fromtimestamp(accessed_time)
            .strftime(GrandPerspectiveWriter.DATE_FORMAT)
            .encode("utf-8")
        )
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.CLOSE_TAG)

        if os.environ.get("PYTEST_CURRENT_TEST"):
            self.file.flush()

    def write_folder_end(self) -> None:
        """Write the end of a folder entry."""

        super().write_folder_end()

        self.file.write(GrandPerspectiveWriter.FOLDER_CLOSE)

        if os.environ.get("PYTEST_CURRENT_TEST"):
            self.file.flush()

    def write_file(
        self, file_name: str, size: int, accessed_time: int, modified_time: int, created_time: int
    ) -> None:
        """Write the start of a file entry."""

        super().write_file(file_name, size, accessed_time, modified_time, created_time)

        self.file.write(GrandPerspectiveWriter.FILE_OPEN)

        self.file.write(GrandPerspectiveWriter.ATTR_NAME)
        self.file.write(file_name.encode("utf-8"))
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_SIZE)
        self.file.write(str(min(self.block_size, size)).encode("utf-8"))
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_CREATED)
        self.file.write(
            datetime.datetime.fromtimestamp(created_time)
            .strftime(GrandPerspectiveWriter.DATE_FORMAT)
            .encode("utf-8")
        )
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_MODIFIED)
        self.file.write(
            datetime.datetime.fromtimestamp(modified_time)
            .strftime(GrandPerspectiveWriter.DATE_FORMAT)
            .encode("utf-8")
        )
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.ATTR_ACCESSED)
        self.file.write(
            datetime.datetime.fromtimestamp(accessed_time)
            .strftime(GrandPerspectiveWriter.DATE_FORMAT)
            .encode("utf-8")
        )
        self.file.write(GrandPerspectiveWriter.ATTR_CLOSE)

        self.file.write(GrandPerspectiveWriter.FILE_CLOSE)

        if os.environ.get("PYTEST_CURRENT_TEST"):
            self.file.flush()



"""A CLI tool for checking disk space."""

import abc
from io import IOBase
import os


class Writer(abc.ABC):
    """A base class for writing disk space results."""

    output_path: str
    block_size: int
    file: IOBase
    file_print_count: int
    current_folder_path = ""

    def __init__(self, output_path: str, file_print_count: int) -> None:
        self.output_path = output_path
        self.block_size = 0
        self.file_print_count = file_print_count
        self.file_count = 0
        self.current_folder_path = ""

    def write_start(
        self,
        root_path: str,
        disk_usage_total: int,
        disk_usage_used: int,
        disk_usage_free: int,
        block_size: int,
    ) -> None:
        """Write the start of the output file."""
        self.file_count = 0

    def write_end(self) -> None:
        """Write the end of the output file."""

    def write_folder_start(
        self, folder_name: str, accessed_time: int, modified_time: int, created_time: int
    ) -> None:
        """Write the start of a folder entry."""
        self.current_folder_path = os.path.join(self.current_folder_path, folder_name)

    def write_folder_end(self) -> None:
        """Write the end of a folder entry."""
        self.current_folder_path = os.path.dirname(self.current_folder_path)

    def write_file(
        self, file_name: str, size: int, accessed_time: int, modified_time: int, created_time: int
    ) -> None:
        """Write the start of a file entry."""
        self.file_count += 1

        if self.file_print_count != 0 and self.file_count % self.file_print_count == 0:
            print(os.path.join(self.current_folder_path, file_name))
